#include "Config.h"
#include "Section.h"
#include "Tagger.h"
#include "Reader.h"
#include "Decoder.h"
#include "TreeDecoder.h"
#include "LanguageModel.h"
#include "Utils.h"
#include "FeatureEnumerator.h"
#include "nlp_string.h"
#include <sys/time.h>
#include <time.h>


void GetTrainingSections (Section* _pRoot, Section_list_t& _lstSections);
void BuildTitleTagLM (LanguageModel &_lmTag, 
					Section_list_t* _plstSections);
void BuildLexicalLM (Section* _pRootSection, 
					LanguageModel &_lmLexical, 
					Section_list_t* _plstSection);


//--------------------------------------------------------------------------------
void TagFromFile (Section* _pRootSection) 
{
    cout << "* Tagging from file..." << endl;

    // Tag text from file
    Tagger tagger;
    tagger.TagFromFile (_pRootSection);
	cout << "  - tagging titles" << endl;
    tagger.TagTitlesFromText (_pRootSection);   
	cout << "  - building np counts" << endl;
	_pRootSection->BuildNPCounts ();
}



//--------------------------------------------------------------------------------
void PreprocessCorpus (Model& _model, 
						Section* _pRootSection, 
						Section_list_t* _plstSections = NULL)
{
    TagFromFile (_pRootSection);
    BuildTitleTagLM (_model.lm_Tag, _plstSections);
    BuildLexicalLM (_pRootSection, _model.lm_Lexical, _plstSections);
}



//--------------------------------------------------------------------------------
void Preprocess (Section* _pRoot)
{
	Section_list_t	lstSubtrees;
	GetTrainingSections (_pRoot, lstSubtrees);

	Section_list_t	lstSections;
	ITERATE_COLLECTION (Section_list_t, lstSubtrees, ite)
	{
		Section* pSection = *ite;
		Utils::GetInOrderSectionList (lstSections, pSection);
	}

	Model model;

	// create language models...								
	cout << "* Preprocessing" << endl;
	PreprocessCorpus (model, _pRoot, &lstSections);
}

