#include <iostream>
#include "Reader.h"
#include "Section.h"


XMLCh*	AlgorithmsReader::px_TitleLiteral = NULL;
XMLCh*	AlgorithmsReader::px_BodyLiteral = NULL;
XMLCh*	AlgorithmsReader::px_SectionLiteral = NULL;
XMLCh*	AlgorithmsReader::px_RootLiteral = NULL;

//=================================================================
Reader::Reader (void)
{
}

Reader::~Reader (void)
{
}


Reader* Reader::CreateReader (const char* _zCorpusName)
{
    if ((0 == strcasecmp (_zCorpusName, DATASET_ALGORITHMS)) ||
		(0 == strcasecmp (_zCorpusName, DATASET_PHYSICS)) ||
		(0 == strcasecmp (_zCorpusName, DATASET_USHISTORY)))
        return new AlgorithmsReader;

	cerr << "Unknown dataset name : '" << _zCorpusName << "'." << endl;
    return NULL;
}


//=================================================================
AlgorithmsReader::AlgorithmsReader (void)
{
}

AlgorithmsReader::~AlgorithmsReader (void)
{
}


//=================================================================
Section* AlgorithmsReader::ParseFile (const char* _zFileName)
{

	XMLPlatformUtils::Initialize ();
	px_TitleLiteral = XMLString::transcode ("title");
	px_BodyLiteral = XMLString::transcode ("body");
	px_SectionLiteral = XMLString::transcode ("section");
	px_RootLiteral = XMLString::transcode ("root");
	
	XercesDOMParser* pParser = new XercesDOMParser ();

	pParser->setValidationScheme (XercesDOMParser::Val_Auto);
	pParser->setDoNamespaces (false);
	pParser->setDoSchema (false);
	pParser->setCreateEntityReferenceNodes (false);

	try
	{
		XMLCh* pxFileName = XMLString::transcode (_zFileName);
		LocalFileInputSource input (pxFileName);
		pParser->parse (input);
		delete pxFileName;

		if (0 != pParser->getErrorCount ())
		{
			std::cerr << "Encountered errors parsing XML file '"
					 << _zFileName << "'. Error count "
					 << pParser->getErrorCount () << std::endl;
			return NULL;
		}
	}
	catch (const DOMException& e)
	{
		std::cerr << "Exception while parsing XML file '"
			 << _zFileName << "'. ";
		if (NULL != e.msg)
		{
			char* zError = XMLString::transcode (e.msg);
			std::cerr << zError << std::endl;
			XMLString::release (&zError);
		}
		else
			std::cerr << "Error code : " << e.code << std::endl;
	}
	catch (const XMLException& e)
	{
		std::cerr << "Exception while parsing XML file '"
			 << _zFileName << "'. Exception : "
			 << e.getMessage () << std::endl;
		return NULL;
	}

	// traverse the XML document & create text section structure.
	DOMDocument* pDoc = pParser->getDocument ();
	if (NULL == pDoc) 
		return NULL;

	Section::i_TotalSections = 0;
	DOMTreeWalker* pWalker = pDoc->createTreeWalker (pDoc, DOMNodeFilter::SHOW_ALL, NULL, true);
	pWalker->firstChild ();
	Section* pRootSection = AlgorithmsReader::Parse (pWalker);
	pRootSection->BuildRootStructures ();

	Section_list_t* pAllSections = pRootSection->GetAllSections ();
	Section* pPreviousSection = NULL;

	Section_list_t::iterator	ite;
	for (ite = pAllSections->begin ();
		ite != pAllSections->end ();
		++ ite)
	{
		Section* pSection = *ite;
		pSection->p_PreviousSection = pPreviousSection;
		pPreviousSection = pSection;
	}
	// deleting the parse is supposed to release all memory allocations associated 	
	// with it - including the DOMDocument & the DOMTreeWalker...					
	delete pParser;

	return pRootSection;
}


//=================================================================
Section* AlgorithmsReader::Parse (DOMTreeWalker* _pWalker)
{
	String sTitle;
	String sBody;

	++ Section::i_TotalSections;
	Section* pSection = new Section;

	DOMNode* pChildNode = _pWalker->firstChild ();
	while (NULL != pChildNode)
	{
		const XMLCh* pxNodeName = pChildNode->getNodeName ();
		if (true == XMLString::equals (pxNodeName, px_TitleLiteral))
		{
			const XMLCh* pxTitleText = pChildNode->getTextContent ();
			char* pTitleText = XMLString::transcode (pxTitleText);
			((string&) sTitle) = pTitleText;
		}
		if (true == XMLString::equals (pxNodeName, px_BodyLiteral))
		{
			const XMLCh* pxBodyText = pChildNode->getTextContent ();
			char* pBodyText = XMLString::transcode (pxBodyText);
			((string&) sBody) = pBodyText;
		}
		if ((true == XMLString::equals (pxNodeName, px_SectionLiteral)) ||
			(true == XMLString::equals (pxNodeName, px_RootLiteral)))
		{
			Section* pChildSection = AlgorithmsReader::Parse (_pWalker);
			pSection->AddChild (pChildSection);
		}

		pChildNode = _pWalker->nextSibling ();
	}
	_pWalker->parentNode ();

	pSection->SetText (sTitle, sBody);
	
	return pSection;
}
                                                                                                                                                                                              





