#include "Section.h"
#include <sys/types.h>
#include <regex.h>
#include <iostream>
#include <fstream>
#include <assert.h>
#include "PorterStemmer.h"
#include "Config.h"
#include "Utils.h"
#include <algorithm>

using namespace std;


Unigram_set_t	Word::set_Stopwords;
Unigram_set_t	Word::set_AuxiliaryWords;
Word*			Word::p_StartWord = NULL;
Word*			Word::p_EndWord = NULL;

Unigram_to_Count_map_t		Section::map_StemToSectionCount;
Unigram_to_Count_map_t		Section::map_WordToSectionCount;
Unigram_to_Count_map_t		Section::map_Stem_to_First_Occurrence_by_Section;
Section_list_t				Section::lst_AllSections;
Section*					Section::p_Root = NULL;
int							Section::i_TotalSections;

//======================================================
Word::Word (const char* _zWord)
{
    s_Word = _zWord;
    Utils::StemWord (s_Stem, _zWord);

    b_IsStopWord = Word::IsStopWord (s_Word);
    b_IsAuxiliaryWord = Word::IsAuxiliaryWord (s_Word);
    b_IsContentWord = ((false == b_IsStopWord) && (false == b_IsAuxiliaryWord));
    b_StartWord = false;
    b_EndWord = false;
}

//======================================================
bool Word::IsStopWord (const String &_sWord) 
{
    if (Word::set_Stopwords.empty ())
	Word::LoadStopwords ();

    return (Word::set_Stopwords.find (_sWord) != Word::set_Stopwords.end ());
}

//======================================================
bool Word::IsAuxiliaryWord (const String &_sWord) 
{
    if (Word::set_AuxiliaryWords.empty ())
	Word::LoadAuxiliaryWords ();

    return (Word::set_AuxiliaryWords.find (_sWord) != Word::set_AuxiliaryWords.end ());
}

//======================================================
void Word::LoadStopwords () 
{
    if (false == Word::set_Stopwords.empty ())
	return;
    
    ifstream file (Config::Get ("STOPWORDS_PATH"));
    if (false == file.is_open ()) 
	{
		cerr << "Failed to open file '" << Config::Get ("STOPWORDS_PATH") << "' to read tagged text." << endl;
		return;
    }

    while (false == file.eof ()) 
	{
		String sLine;
		getline (file, sLine);
		Word::set_Stopwords.insert (sLine);
    }
}

//======================================================
void Word::LoadAuxiliaryWords () 
{
    if (false == Word::set_AuxiliaryWords.empty ())
		return;
    
    ifstream file (Config::Get ("AUXILIARY_WORDS_PATH"));
    if (false == file.is_open ()) 
	{
		cerr << "Failed to open file '" << Config::Get ("AUXILIARY_WORDS_PATH") << "' to read tagged text." << endl;
		return;
    }

    while (false == file.eof ()) 
	{
		String sLine;
		getline (file, sLine);
		Word::set_AuxiliaryWords.insert (sLine);
    }
}

//======================================================
Word* Word::GetStartWord (void) 
{
    if (p_StartWord == NULL) 
    {
        p_StartWord = new Word ("#start#");
        p_StartWord->s_POSTag = "#start#";

		p_StartWord->b_StartWord = true;

		p_StartWord->b_EndWord = false;
		p_StartWord->b_IsStopWord = false;
		p_StartWord->b_IsAuxiliaryWord = false;
		p_StartWord->b_IsContentWord = false;
    }
    return p_StartWord;
}


//======================================================
Word* Word::GetEndWord (void) 
{
    if (p_EndWord == NULL) 
    {
        p_EndWord = new Word ("#end#");
        p_EndWord->s_POSTag = "#end#";

		p_EndWord->b_EndWord = true;

		p_EndWord->b_StartWord = false;
		p_EndWord->b_IsStopWord = false;
		p_EndWord->b_IsAuxiliaryWord = false;
		p_EndWord->b_IsContentWord = false;
    }
    return p_EndWord;
}

//======================================================
bool Word::operator== (const Word& _rRight) const
{
    return ((_rRight.s_Word == s_Word) && (_rRight.s_POSTag == s_POSTag));
}

bool Word::operator!= (const Word& _rRight) const
{
    return ((_rRight.s_Word != s_Word) || (_rRight.s_POSTag != s_POSTag));
}


//======================================================
String Word::GetWordText (Word& _rWord)
{
	return _rWord.s_Word;
}

String Word::GetTagText (Word& _rWord)
{
	return _rWord.s_POSTag;
}

String Word::GetWordTagText (Word& _rWord)
{
	return (_rWord.s_Word + '_' + _rWord.s_POSTag);
}


//======================================================
ostream& operator<< (ostream& _ofs, const Word& _rWord)
{
	_ofs << _rWord.s_Word << '/' << _rWord.s_POSTag << ' ';
	return _ofs;
}

ostream& operator<< (ostream& _ofs, const Word_dq_t& _dqWords)
{
	ITERATE_COLLECTION (Word_dq_t, ((Word_dq_t&)_dqWords), ite)
	{
		Word* pWord = *ite;
		_ofs << pWord->s_Word << '/' << pWord->s_POSTag << ' ';
	}
	return _ofs;
}

ostream& operator<< (ostream& _ofs, const Word_list_t& _lstWords)
{
	ITERATE_COLLECTION (Word_list_t, ((Word_list_t&)_lstWords), ite)
	{
		Word* pWord = *ite;
		_ofs << pWord->s_Word << '/' << pWord->s_POSTag << ' ';
	}
	return _ofs;
}


//======================================================
Sentence::Sentence (const char* _zText)
{
	s_Text = _zText;
	
	Utils::CleanText (s_Text);
	Utils::TokenizeText (s_Text);

	list <String>	lstTokens;
	Utils::SplitIntoWords (s_Text, lstTokens);

	list <String>::iterator	ite;
	for (ite = lstTokens.begin ();
		ite != lstTokens.end ();
		++ ite)
	{
		ite->LowerCase();

	    Word* pWord = new Word (*ite);
	    lst_Words.push_back (pWord);
	    set_Stems.insert (pWord->s_Stem);
	}
}


//====================================================== <<<<<<<<<<<<<<
String Sentence::GetTaggedText (void)
{
	String sTaggedText;
	String sLowerText = s_Text;

	sLowerText.LowerCase ();

	Unigram_list_t lstLowerTokens;
	sLowerText.Split (lstLowerTokens);

	Word_list_t::iterator iteWord = lst_Words.begin ();
	unsigned int iTokenIndex = 0;
	ITERATE_COLLECTION (Unigram_list_t, lstLowerTokens, iteLowerToken)  
	{
	    String sLowerToken (*iteLowerToken);
		String sToken (sLowerToken);

		if ((sToken == "-lrb-") || (sToken == "-rrb-") ||
			(sToken == "-lsb-") || (sToken == "-rsb-") ||
			(sToken == "-lcb-") || (sToken == "-rcb-"))
			sToken.UpperCase ();

	    if (0 != iTokenIndex)
			sTaggedText << ' ';
	    sTaggedText << sToken;
	    
		Word* pWord = *iteWord;
		if (iteWord != lst_Words.end () && sLowerToken == pWord->s_Word) 
		{
			sTaggedText << "_" << pWord->s_POSTag;
			++iteWord;
		}
	    
	    ++iTokenIndex;
	}

	return sTaggedText;
}


//======================================================
Word_list_t& Sentence::GetWords (void)  
{ 
	return lst_Words; 
}


//======================================================
const char* Sentence::GetText (void) 
{ 
	return s_Text; 
}


//======================================================
Section::Section ( void )
{
    p_Title = NULL;
    p_Parent = NULL;
    p_PreviousSection = NULL;
    p_NextSibling = NULL;
    p_PreviousSibling = NULL;
    i_Depth = -1;
 }

//======================================================
Section::Section (const char* _zTitle, const char* _zBody)
{
    p_Title = NULL;
    p_Parent = NULL;
    p_PreviousSection = NULL;
    p_NextSibling = NULL;
    p_PreviousSibling = NULL;
    i_Depth = -1;

    Section::SetText (_zTitle, _zBody);
}

//======================================================
Section::~Section (void)
{
}


//======================================================
void Section::SetText (const char* _zTitle, const char* _zBody)
{
    p_Title = new Sentence (_zTitle);

    String sBody (_zBody);
	Utils::CleanText (sBody);

    list <String> lstSentenceText;
    Utils::SplitIntoSentences (sBody, lstSentenceText);
    
    ITERATE_COLLECTION (list <String>, lstSentenceText, ite)
	lst_Sentences.push_back (new Sentence (*ite));

    i_TotalUnigrams = 0;   
    long lSentencePosition = 0;
    ITERATE_COLLECTION (Sentence_list_t, lst_Sentences, iteSentence)
	{
	    Sentence* pSentence = *iteSentence;
	    
	    long lWordPosition = 0;
	    ITERATE_COLLECTION (Word_list_t, pSentence->lst_Words, iteWord)
		{
		    Word* pWord = *iteWord;
		    
		    map_Stem_to_First_Occurrence_by_Word.insert (make_pair (pWord->s_Stem, lWordPosition));
		    map_Stem_to_First_Occurrence_by_Sentence.insert (make_pair (pWord->s_Stem, lSentencePosition));
		    
		    Unigram_to_Count_map_t::iterator	iteCount;
		    
		    // stem counts...					
		    iteCount = map_Stem_to_Count.find (pWord->s_Stem);
		    if (iteCount == map_Stem_to_Count.end ())
				map_Stem_to_Count.insert (make_pair (pWord->s_Stem, 1));
		    else
				++ iteCount->second;
		    
		    // word counts...					
		    iteCount = map_Word_to_Count.find (pWord->s_Word);
		    if (iteCount == map_Word_to_Count.end ())
				map_Word_to_Count.insert (make_pair (pWord->s_Word, 1));
		    else
				++ iteCount->second;
		    
		    ++ lWordPosition;
		    ++i_TotalUnigrams;
		}
	    ++ lSentencePosition;
	}
}


//======================================================
Word_list_t Section::GetBodyWords (void)
{
    Word_list_t lstBodyWords;

    ITERATE_COLLECTION (Sentence_list_t, lst_Sentences, iteSentence)
	{
	    Sentence* pSentence = *iteSentence;
	    
	    ITERATE_COLLECTION (Word_list_t, pSentence->lst_Words, iteWord)
		lstBodyWords.push_back (*iteWord);
	}
	 
    return lstBodyWords;
}   


//-------------------------------------------------------------------
void Section::CreateUniqueWordList (void)
{
	typedef map <String, Word*>	UniqueWord_map_t;
	UniqueWord_map_t	mapUniqueWords;

	ITERATE_COLLECTION (Sentence_list_t, lst_Sentences, iteSentence)
	{
		Sentence* pSentence = *iteSentence;
		ITERATE_COLLECTION (Word_list_t, pSentence->lst_Words, iteWord)
		{
			Word* pWord = *iteWord;

			String sKey (pWord->s_Word);
			sKey << '|' << pWord->s_POSTag;

			mapUniqueWords.insert (make_pair (sKey, pWord));
		}
	}

	ITERATE_COLLECTION (UniqueWord_map_t, mapUniqueWords, ite)
		dq_UniqueWords.push_back (ite->second);

	mapUniqueWords.clear ();
}


//-------------------------------------------------------------------
Unigram_list_t Section::CapitalizeTitle (Word_list_t _lstTitleWords) 
{
	// Build Capitalization Map
    typedef map <String, int> Word_to_CapitalCount_map_t;
    typedef map <String, Word_to_CapitalCount_map_t> Word_to_CapitalWordAndCountMap_map_t;
    Word_to_CapitalWordAndCountMap_map_t map_Word_to_Capital;

    ITERATE_COLLECTION (Sentence_list_t, lst_Sentences, iteSentence) 
	{
		Unigram_list_t lstSentenceTokens;
		(*iteSentence)->s_Text.Split (lstSentenceTokens);

		ITERATE_COLLECTION (Unigram_list_t, lstSentenceTokens, iteWord) 
		{
			if ((lstSentenceTokens.begin () != iteWord) || 
				(lstSentenceTokens.begin () == iteWord && iteWord->IsUpper ())) 
			{
				String sLowerWord = *iteWord;
				sLowerWord.LowerCase ();

				if (map_Word_to_Capital.find (sLowerWord) == map_Word_to_Capital.end ())
				{
					Word_to_CapitalCount_map_t map_CapitalWord_to_Count; 
					map_Word_to_Capital.insert (make_pair (sLowerWord, map_CapitalWord_to_Count));
				}
		
				Word_to_CapitalWordAndCountMap_map_t::iterator	iteCapitalCount;
				iteCapitalCount = map_Word_to_Capital.find (sLowerWord);
				if (iteCapitalCount->second.find (*iteWord) == iteCapitalCount->second.end ())
					map_Word_to_Capital [sLowerWord] [*iteWord] = 0;

				++map_Word_to_Capital [sLowerWord] [*iteWord];
			}
		}
    }

    // Find max case for each word
    typedef map <String, String> map_Word_to_CapWord_t;
    map_Word_to_CapWord_t map_Word_to_CapWord;
    ITERATE_COLLECTION (Word_to_CapitalWordAndCountMap_map_t, map_Word_to_Capital, iteWordCap) 
	{
		String sWord = iteWordCap->first;
		String sMaxCase;
		int iMaxCaseCount = -1;
		ITERATE_COLLECTION (Word_to_CapitalCount_map_t, map_Word_to_Capital [sWord], iteCapCount) 
		{
			String sCasedWord = iteCapCount->first;
			int iCount = iteCapCount->second;
			
			if (iMaxCaseCount <= iCount)
			{
				iMaxCaseCount = iCount;
				sMaxCase = sCasedWord;
			}
		}
		map_Word_to_CapWord.insert (make_pair (sWord, sMaxCase));
    }

    // Case the title
    Unigram_list_t lstCapitalizedTitle;
    bool bFirstWord = true;
    ITERATE_COLLECTION (Word_list_t, _lstTitleWords, iteTitleWord) 
	{
		String sTitleWord = (*iteTitleWord)->s_Word;
		if (*iteTitleWord == Word::GetStartWord ()) 
		{
			lstCapitalizedTitle.push_back (sTitleWord);
			continue;
		}


		String sCapitalizedWord;
		if (map_Word_to_CapWord.find (sTitleWord) != map_Word_to_CapWord.end ())
			sCapitalizedWord = (const char*)map_Word_to_CapWord [sTitleWord];
		else
			sCapitalizedWord = sTitleWord;


		if (true == bFirstWord) 
		{
			if (sCapitalizedWord.IsLower ()) 
				sCapitalizedWord.Capitalize ();

			lstCapitalizedTitle.push_back (sCapitalizedWord);
			bFirstWord = false;
		} 
		else
			lstCapitalizedTitle.push_back (sCapitalizedWord);
	}
    
    return lstCapitalizedTitle;
}


//-------------------------------------------------------------------
bool Section::IsRoot (void) 
{
    return p_Parent == NULL;
}


//-------------------------------------------------------------------
void Section::GetDescendants (Section_list_t* _rlstSections)
{
	_rlstSections->push_back (this);

	ITERATE_COLLECTION (Section_list_t, lst_Children, ite) 
	{
		Section* pSection = *ite;
		pSection->GetDescendants (_rlstSections);
	}
}


//-------------------------------------------------------------------
Section* Section::GetRoot (void) 
{
    return Section::p_Root;
}


//-------------------------------------------------------------------
Section_list_t* Section::GetAllSections (void) 
{
	return &lst_AllSections;
}


//-------------------------------------------------------------------
void Section::AddChild (Section* _pChild)
{
    _pChild->p_Parent = this;

	_pChild->p_NextSibling = NULL;
	if (lst_Children.empty ())
		_pChild->p_PreviousSibling = NULL;
	else
	{
		Section* pPreviousSibling = lst_Children.back ();
		pPreviousSibling->p_NextSibling = _pChild;
		_pChild->p_PreviousSibling = pPreviousSibling;
	}

    lst_Children.push_back (_pChild);
}


//-------------------------------------------------------------------
void Section::BuildRootStructures (void)
{
    Section::p_Root = this;
    while (NULL != Section::p_Root->p_Parent)
		Section::p_Root = Section::p_Root->p_Parent;

	Section::p_Root->GetDescendants (&lst_AllSections);
	Section::p_Root->BuildFirstSectionWithStemMap ();
	BuildUnigramToSectionCounts ();
}


//-------------------------------------------------------------------
void Section::BuildUnigramToSectionCounts (void)
{
	Section_list_t* plstAllSection = GetAllSections ();
    ITERATE_COLLECTION_POINTER (Section_list_t, plstAllSection, iteSection) 
	{
		Section* pSection = *iteSection;
		ITERATE_COLLECTION (Unigram_to_Count_map_t, pSection->map_Stem_to_Count, iteStem)
		{
			Unigram_to_Count_map_t::iterator	ite;
			ite = map_StemToSectionCount.find (iteStem->first);
			if (ite == map_StemToSectionCount.end ())
				map_StemToSectionCount.insert (make_pair (iteStem->first, 1));
			else
				++ ite->second;
		}
		ITERATE_COLLECTION (Unigram_to_Count_map_t, pSection->map_Word_to_Count, iteWord)
		{
			Unigram_to_Count_map_t::iterator	ite;
			ite = map_WordToSectionCount.find (iteWord->first);
			if (ite == map_WordToSectionCount.end ())
				map_WordToSectionCount.insert (make_pair (iteWord->first, 1));
			else
				++ ite->second;
		}
    }
}


//-------------------------------------------------------------------
void Section::BuildFirstSectionWithStemMap (void) 
{
    if (false == IsRoot ()) 
	{
		cerr << "BuildFirstSectionWithStemMap: This should only be called on the root node" << endl;
		throw;
    }

    int iSectionNumber = 0;
    ITERATE_COLLECTION_POINTER (Section_list_t, GetAllSections (), iteSection) 
	{
		ITERATE_COLLECTION (Unigram_to_Count_map_t, (*iteSection)->map_Stem_to_First_Occurrence_by_Word, iteStem) 
		{
			String sStem = iteStem->first;
			if (map_Stem_to_First_Occurrence_by_Section.find (sStem) == map_Stem_to_First_Occurrence_by_Section.end ())
				map_Stem_to_First_Occurrence_by_Section.insert (make_pair (sStem, iSectionNumber));
		}
		++iSectionNumber;
    }
}


//-------------------------------------------------------------------
void Section::BuildNPCounts (void) 
{
    if (false == map_NP_to_Corpus_Count.empty ())
		return;

    if (false == IsRoot ()) 
	{
		cerr << "BuildNPCounts: This should only be called on the root section" << endl;
		throw;
    }

	ITERATE_COLLECTION_POINTER (Section_list_t, GetAllSections (), iteSection) 
	{
		Section* pSection = *iteSection;
		ITERATE_COLLECTION (Sentence_list_t, pSection->lst_Sentences, iteSentence) 
		{
			Sentence* pSentence = *iteSentence;

			String sText (pSentence->s_Text);
			sText.LowerCase ();
			Ngram_list_t lstNgrams;

			Utils::ExtractNounPhrases (lstNgrams, &sText, &(pSentence->lst_Words));
			ITERATE_COLLECTION (Ngram_list_t, lstNgrams, iteNgram) 
			{
				String sNgram;
				Utils::WordListToNgramString (sNgram, (*iteNgram));

				Ngram_to_Count_map_t::iterator iteCount;
				iteCount = pSection->map_NP_to_Count.find (sNgram);
				if (iteCount == pSection->map_NP_to_Count.end ())
					pSection->map_NP_to_Count.insert (make_pair (sNgram, 1));
				else
					++ iteCount->second;
			}
			Utils::DeleteNgramList( lstNgrams );
		}

		Ngram_to_Count_map_t& mapNPCounts = pSection->map_NP_to_Count;
		ITERATE_COLLECTION (Ngram_to_Count_map_t, mapNPCounts, iteNPCount) 
		{
			const char* zNgram = iteNPCount->first;
			int iNgramCount = iteNPCount->second;

			Ngram_to_Count_map_t::iterator	iteCount;
			iteCount = map_NP_to_Corpus_Count.find (zNgram);
			if (iteCount == map_NP_to_Corpus_Count.end ())
				map_NP_to_Corpus_Count.insert (make_pair (zNgram, iNgramCount));
			else
				iteCount->second += iNgramCount;
		}
	}
}


//-------------------------------------------------------------------
bool Section::HasUnigram (const char* _zUnigram, bool _bUseStem) 
{
	if (true == _bUseStem)
		return (map_Stem_to_Count.find (_zUnigram) != map_Stem_to_Count.end ());
	else
		return (map_Word_to_Count.find (_zUnigram) != map_Word_to_Count.end ());
}


//-------------------------------------------------------------------
int Section::GetUnigramOccurrence (const char* _zUnigram, bool _bUseStem) 
{
	Unigram_to_Count_map_t::iterator	ite;
    if (_bUseStem)
	{
		ite = map_Stem_to_Count.find (_zUnigram);
		if (ite == map_Stem_to_Count.end ())
			return 0;

		return ite->second;
	}
    else
	{
		ite = map_Word_to_Count.find (_zUnigram);
		if (ite == map_Word_to_Count.end ())
			return 0;

		return ite->second;
	}
}


//-------------------------------------------------------------------
int Section::GetNumberOfSectionsWithUnigram (const char* _zUnigram, bool _bUseStem) 
{
	if (true == _bUseStem)
		return map_StemToSectionCount [_zUnigram];
	else
		return map_WordToSectionCount [_zUnigram];
}


//-------------------------------------------------------------------
int Section::GetTotalUnigramOccurrences (void)
{
    return i_TotalUnigrams;
}


//-------------------------------------------------------------------
long Section::GetFirstOccurrenceByWord (const char* _zStem) 
{
    return map_Stem_to_First_Occurrence_by_Word [_zStem];
}


//-------------------------------------------------------------------
long Section::GetFirstOccurrenceBySentence (const char* _zStem) 
{
    return map_Stem_to_First_Occurrence_by_Sentence [_zStem];
}

//-------------------------------------------------------------------
long Section::GetFirstOccurrenceBySection (const char* _zStem) 
{
    if (false == IsRoot ()) 
	{
		cerr << "GetFirstOccurrenceBySection: This method can only be called on the root section" << endl;
		throw;
    }
    // Note: If the stem is not found in corpus, we return the last section
	Unigram_to_Count_map_t::iterator ite;
	ite = map_Stem_to_First_Occurrence_by_Section.find (_zStem);
	if (ite != map_Stem_to_First_Occurrence_by_Section.end ())
		return ite->second;

	return GetAllSections ()->size ();
} 


//-------------------------------------------------------------------
int Section::GetNPCount (const char* _zNP) 
{
	Ngram_to_Count_map_t::iterator	ite;
	ite = map_NP_to_Count.find (_zNP);
	if (ite == map_NP_to_Count.end ())
		return 0;

	return ite->second;
}


//-------------------------------------------------------------------
set<String> Section::GetBodyNPs (void) 
{
    set<String> set_NPs;
    ITERATE_COLLECTION (Ngram_to_Count_map_t, map_NP_to_Count, iteNPCount) 
		set_NPs.insert (iteNPCount->first);
 
    return set_NPs;
}


//-------------------------------------------------------------------
int Section::GetCorpusNPCount (const char* _zNP) 
{
    if (false == IsRoot ()) 
	{
		cerr << "GetCorpusNPCount: This method can only be called on the root section" << endl;
		throw;
    }

	Ngram_to_Count_map_t::iterator	ite;
	ite = map_NP_to_Corpus_Count.find (_zNP);
	if (ite == map_NP_to_Corpus_Count.end ())
		return 0;

	return ite->second;
}


//-------------------------------------------------------------------
bool Section::PreviousSiblingHasStem (const char* _zStem) 
{
	if (NULL == p_PreviousSibling)
		return false;
	return p_PreviousSibling->HasUnigram (_zStem, true);
}

//-------------------------------------------------------------------
bool Section::NextSiblingHasStem (const char* _zStem) 
{
	if (NULL == p_NextSibling)
		return false;
	return p_NextSibling->HasUnigram (_zStem, true);
}


//-------------------------------------------------------------------
Unigram_to_Count_map_t* Section::GetAllWords (void)
{
	return &map_WordToSectionCount;
}

//-------------------------------------------------------------------
int Section::GetDepth (void)
{
    if (-1 != i_Depth)
		return i_Depth;
    
    Section *pSection = this;
    while (NULL != pSection->p_Parent) 
	{
		pSection = pSection->p_Parent;
		++ i_Depth;
    }
    
    return i_Depth;
}
